package org.openmuc.jdlms;

import java.io.IOException;
import java.security.InvalidKeyException;
import java.security.Key;
import java.security.NoSuchAlgorithmException;
import java.security.SecureRandom;
import java.util.Arrays;
import java.util.List;

import javax.crypto.BadPaddingException;
import javax.crypto.Cipher;
import javax.crypto.IllegalBlockSizeException;
import javax.crypto.NoSuchPaddingException;
import javax.crypto.spec.SecretKeySpec;

import org.bouncycastle.crypto.engines.AESFastEngine;
import org.bouncycastle.crypto.engines.RFC3394WrapEngine;
import org.bouncycastle.crypto.params.KeyParameter;
import org.openmuc.jdlms.datatypes.DataObject;
import org.openmuc.jdlms.interfaceclass.method.SecuritySetupMethod;

public class SecurityUtils {

	public enum KeyId {
		GLOBAL_UNICAST_ENCRYPTION_KEY(0),
		GLOBAL_BROADCAST_ENCRYPTION_KEY(1),
		AUTHENTICATION_KEY(2);

		private final int id;

		private KeyId(int id) {
			this.id = id;
		}

		public int keyId() {
			return id;
		}

		public String keyName() {
			return name();
		}
	}

	public enum SecurityPolicy {
		NOTHING(0),
		AUTHENTICATED(1),
		ENCRYPTED(2),
		AUTHENTICATED_AND_ENCRYPTED(3);

		private final int id;

		private SecurityPolicy(int id) {
			this.id = id;
		}

		public int keyId() {
			return id;
		}

		public String keyName() {
			return name();
		}
	}

	/**
	 * Returns the method parameter for updating a single key of a dlms server.
	 * 
	 * @param masterKey
	 *            the master key, also known as kek
	 * @param globalKey
	 *            the new key to update to the dlms server
	 * @param keyId
	 *            the type of key to update
	 * @return return MethodParameter for global key transfer
	 * @throws IOException
	 *             throws IOException
	 */
	public static MethodParameter globalKeyTransfer(byte[] masterKey, byte[] globalKey, KeyId keyId)
			throws IOException {
		final byte instance = 0; // current instance
		ObisCode obisCode = new ObisCode(0, 0, 43, 0, instance, 255);

		final byte[] wrappedKey = aesRFC3394KeyWrap(masterKey, globalKey);

		List<DataObject> keyDataList = Arrays.asList(DataObject.newEnumerateData(keyId.id),
				DataObject.newOctetStringData(wrappedKey));
		DataObject keyData = DataObject.newStructureData(keyDataList);
		DataObject methodParameter = DataObject.newArrayData(Arrays.asList(keyData));

		MethodParameter param = new MethodParameter(SecuritySetupMethod.GLOBALE_KEY_TRANSFER, obisCode,
				methodParameter);

		return param;
	}

	/**
	 * Sets the security policy method parameter
	 * 
	 * @param securityPolicy
	 *            the security policy to set for
	 * @return return MethodParameter for security policy
	 * @throws IOException
	 *             throws IOException
	 */
	public static MethodParameter securityActivate(SecurityPolicy securityPolicy) throws IOException {
		final byte instance = 0; // current instance
		ObisCode obisCode = new ObisCode(0, 0, 43, 0, instance, 255);

		MethodParameter param = new MethodParameter(SecuritySetupMethod.SECURITY_ACTIVATE, obisCode,
				DataObject.newEnumerateData(securityPolicy.id));

		return param;
	}

	/**
	 * Encrypts a byte array with a master key with the algorithm AES in mode CBC and no padding.
	 * 
	 * @param masterKey
	 *            the master key for encryption the bytesToCypher
	 * @param bytesToCipher
	 *            the bytes to cipher
	 * @return returns the bytesToCipher encrypted
	 * @throws NoSuchAlgorithmException
	 *             throws a NoSuchAlgorithmException
	 * @throws NoSuchPaddingException
	 *             throws a NoSuchPaddingException
	 * @throws InvalidKeyException
	 *             throws a InvalidKeyException
	 * @throws IllegalBlockSizeException
	 *             throws a IllegalBlockSizeException
	 * @throws BadPaddingException
	 *             throws a BadPaddingException
	 */
	public static byte[] aes128Ciphering(byte[] masterKey, byte[] bytesToCipher) throws NoSuchAlgorithmException,
			NoSuchPaddingException, InvalidKeyException, IllegalBlockSizeException, BadPaddingException {
		Key secretkeySpec = new SecretKeySpec(masterKey, "AES");
		Cipher cipher = Cipher.getInstance("AES/CBC/NoPadding");
		cipher.init(Cipher.ENCRYPT_MODE, secretkeySpec);
		byte[] encryptedKey = cipher.doFinal(bytesToCipher);

		return encryptedKey;
	}

	/**
	 * This function wraps a key with a kek (key encryption key)
	 * 
	 * @param kek
	 *            the key encryption key for wrapping the key
	 * @param key
	 *            the key to wrap
	 * @return returns a with kek wrapped key
	 */
	public static byte[] aesRFC3394KeyWrap(byte[] kek, byte[] key) {
		RFC3394WrapEngine rfc3394WrapEngine = new RFC3394WrapEngine(new AESFastEngine());
		rfc3394WrapEngine.init(true, new KeyParameter(kek));

		byte[] wrappedKey = rfc3394WrapEngine.wrap(key, 0, key.length);

		return wrappedKey;
	}

	/**
	 * This function unwraps a wrapped key with the kek (key encryption key)
	 * 
	 * @param kek
	 *            the key encryption key for unwrapping the wrapped key
	 * @param wrappedKey
	 *            the wrapped key to unwrap
	 * @return returns a unwrapped key
	 * @throws InvalidCipherTextException
	 *             will thrown if something unexpected is in the wrappedKey
	 */
	// public static byte[] aesRFC3394KeyUnwrap(byte[] kek, byte[] wrappedKey) throws InvalidCipherTextException {
	// RFC3394WrapEngine rfc3394WrapEngine = new RFC3394WrapEngine(new AESFastEngine());
	// rfc3394WrapEngine.init(false, new KeyParameter(wrappedKey));
	//
	// byte[] key = rfc3394WrapEngine.unwrap(kek, 0, kek.length);
	// return key;
	// }

	/**
	 * Generates a random AES 128 key
	 * 
	 * @return returns a random AES 128 key
	 */
	public static byte[] generateAES128Key() {
		byte[] key = new byte[16];

		SecureRandom secureRandom = new SecureRandom();
		secureRandom.nextBytes(key);

		return key;
	}

}
